'use strict';

var lodash = require('lodash');

// Returns a lodash chain that calls .value()
// automatically after the first .method()
//
// It also returns a promise or value
//
// For example:
// lowChain(_, array, save).method()
//
// is the same as:
// _.chain(array).method().value()
function lowChain(_, array, save) {
  var chain = _.chain(array);

  _.functions(chain).forEach(function (method) {
    chain[method] = _.flow(chain[method], function (arg) {
      var v = arg.value ? arg.value() : arg;
      var s = save();

      if (s) return s.then(function () {
        return Promise.resolve(v);
      });
      return v;
    });
  });

  return chain;
}

function low(source) {
  var options = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];
  var writeOnChange = arguments.length <= 2 || arguments[2] === undefined ? true : arguments[2];

  // Create a fresh copy of lodash
  var _ = lodash.runInContext();

  if (source) {
    if (options.storage) {
      (function () {
        var storage = options.storage;

        db.read = storage.read;
        db.write = storage.write ? function () {
          var dest = arguments.length <= 0 || arguments[0] === undefined ? source : arguments[0];
          return storage.write(dest, db.object, db.serialize);
        } : null;
      })();
    }

    options.format && Object.assign(db, options.format);
  }

  // Modify value function to call save before returning result
  _.prototype.value = _.wrap(_.prototype.value, function (value) {
    var v = value.apply(this);
    var s = _save();

    if (s) return s.then(function () {
      return Promise.resolve(v);
    });
    return v;
  });

  // Init db.object checksum
  var checksum = JSON.stringify(db.object);

  // Return a promise or nothing in sync mode or if the database hasn't changed
  function _save() {
    if (db.source && db.write && writeOnChange) {
      var str = JSON.stringify(db.object);

      if (str !== checksum) {
        checksum = str;
        return db.write(db.source, db.object);
      }
    }
  }

  function db(key) {
    var array = db.object[key] = db.object[key] || [];
    var short = lowChain(_, array, _save);
    short.chain = function () {
      return _.chain(array);
    };
    return short;
  }

  // Expose
  db._ = _;
  db.object = {};
  db.source = source;

  // Init
  db.source && db.read && db.read(source, db.deserialize);

  return db;
}

module.exports = low;